package com.softslate.commerce.administrator.core;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Map;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.Globals;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.administrator.Administrator;
import com.softslate.commerce.businessobjects.administrator.AdministratorProcessor;
import com.softslate.commerce.businessobjects.core.BusinessObjectFactory;
import com.softslate.commerce.businessobjects.core.Settings;
import com.softslate.commerce.businessobjects.core.User;
import com.softslate.commerce.customer.core.ActionUtils;
import com.softslate.commerce.customer.core.CheckEERequestProcessor;

/**
 * Subclass of <code>BaseRequestProcessor</code> that negotiates each
 * application request for the customer interface. The
 * "/WEB-INF/conf/administrator/core/struts-config.xml" file identifies this
 * class as the customer application's &lt;controller&gt;.
 * 
 * @author David Tobey
 */
public class AdministratorRequestProcessor extends CheckEERequestProcessor {

	static Log log = LogFactory.getLog(AdministratorRequestProcessor.class);

	public static final String errorPath = "/administrator/Error.do";

	public static final String loginPath = "/administrator/Login.do";

	public static final String loginFormPath = "/administrator/LoginForm.do";

	public static final String changePasswordPath = "/administrator/AdministratorChangePasswordAndLogin.do";

	public static final String changePasswordFormPath = "/administrator/AdministratorChangePasswordForm.do";

	public static final String apiErrorPath = "/WEB-INF/templates/administrator/default/json/justErrorsAndMessages.jsp";

	public static final String customLayoutDirectory = "custom";

	/**
	 * General-purpose preprocessing hook. Loads settings into the request and
	 * session scopes.
	 * 
	 * @param request
	 *            The servlet request we are processing
	 * @param response
	 *            The servlet response we are creating
	 */
	protected boolean processPreprocess(HttpServletRequest request, HttpServletResponse response) {
		boolean cont = super.processPreprocess(request, response);
		if (!cont) {
			return false;
		}

		try {

			// Forbid non-secure requests if called for by settings
			if (forbiddenBecauseRequestNotSecure(request, response)) {
				return false;
			}

			// Set the length of the administrator's session. The number of minutes of inactivity before the session is
			// killed on the server. Defaults to container default (15-20 minutes for Tomcat) or the config in web.xml.
			Settings settings = (Settings) servlet.getServletContext().getAttribute("settings");
			if (settings != null) {
				String sessionTimeoutAdministrators = (String) settings.getValue("sessionTimeoutAdministrators");
				if (StringUtils.isNotBlank(sessionTimeoutAdministrators) && !sessionTimeoutAdministrators.equals("0")
						&& !(Boolean) request.getSession().getAttribute("sessionTimeoutDefined")) {
					synchronized (request.getSession()) {
						request.getSession().setMaxInactiveInterval(Integer.parseInt(sessionTimeoutAdministrators));
						request.getSession().setAttribute("sessionTimeoutDefined", Boolean.TRUE);
					}
				}
			}

			// Set the custom layout directory to "custom" for custom templates
			synchronized (request.getSession()) {
				request.getSession().setAttribute("adminCurrentLayout", customLayoutDirectory);
			}

			// Allowed requests
			if (authenticateApiRequest(request) || isAdministratorLoggedIn(request) || isLoginRequest(request)
					|| isLegacyApiRequest(request, response)) {
				// Call a template method for hooks.
				return sscOnAdministratorPreRequestEnd(request, response);
			} else {
				if (isApiRequest(request)) {
					redirectToApiInvalidLogin(request, response);
				} else {
					redirectToLoginFormPreservingOriginalRequest(request, response);
				}
				return false;
			}
		} catch (Exception se) {
			log.error("Exception processing admin request: ", se);
			return (false);
		}
	}

	protected boolean forbiddenBecauseRequestNotSecure(HttpServletRequest req, HttpServletResponse res)
			throws ServletException, IOException {
		boolean ret = false;
		Settings settings = (Settings) req.getSession().getServletContext().getAttribute("settings");
		if (settings != null && settings.getValue("rejectNonSecureAdminRequests") != null
				&& settings.getValue("rejectNonSecureAdminRequests").equals("1") && !req.isSecure()) {
			if (log.isDebugEnabled()) {
				log.debug("Rejecting admin request because it is not secure.");
			}
			forbidRequest(req, res);
			ret = true;
		}
		return ret;
	}

	protected boolean isLoginRequest(HttpServletRequest req) {
		boolean ret = req.getServletPath().endsWith(loginFormPath) || req.getServletPath().endsWith(loginPath)
				|| req.getServletPath().endsWith(changePasswordPath)
				|| req.getServletPath().endsWith(changePasswordFormPath);
		return ret;
	}

	protected boolean isApiRequest(HttpServletRequest req) {
		boolean ret = Boolean.TRUE.equals(req.getAttribute("isApiRequest"));
		return ret;
	}

	protected boolean isAdministratorLoggedIn(HttpServletRequest req) {
		User user = (User) req.getSession().getAttribute("user");
		Administrator administrator = user.getAdministrator();
		boolean ret = administrator != null;
		if (ret && log.isDebugEnabled()) {
			log.debug(req.getRemoteUser() + " logged in.");
		}
		return ret;
	}

	protected boolean authenticateApiRequest(HttpServletRequest req) throws Exception {
		boolean isAuthentic = false;
		if (req.getAttribute("apiOutputFormat") != null) {
			isAuthentic = true;
		} else {
			String apiLoginID = req.getParameter("apiLoginID");
			String apiAuthenticationToken = req.getParameter("apiAuthenticationToken");
			String apiOutputFormat = "json";
			if (apiLoginID == null) {
				apiLoginID = req.getHeader("apiLoginID");
				apiAuthenticationToken = req.getHeader("apiAuthenticationToken");
			}
			if (apiLoginID != null && apiAuthenticationToken != null) {
				BusinessObjectFactory bof = (BusinessObjectFactory) req.getAttribute("businessObjectFactory");
				AdministratorProcessor administratorProcessor =  bof
						.createObject(AdministratorProcessor.class);
				Map result = administratorProcessor.processApiLogin(apiLoginID, apiAuthenticationToken);
				if (result != null && result.get("resultCode") != null && result.get("resultCode").equals("0")) {
					req.setAttribute("apiOutputFormat", apiOutputFormat);
					isAuthentic = true;
				}
			}
		}
		return isAuthentic;
	}

	protected void forbidRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException,
			IOException {
		ActionMessages errors = new ActionMessages();
		errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.administratorForbidden"));
		request.setAttribute(Globals.ERROR_KEY, errors);
		response.setStatus(HttpServletResponse.SC_FORBIDDEN);
		RequestDispatcher rd = null;
		if (isApiRequest(request)) {
			rd = request.getRequestDispatcher(apiErrorPath);
		} else {
			rd = request.getRequestDispatcher(errorPath);
		}
		rd.forward(request, response);
	}

	protected void redirectToApiInvalidLogin(HttpServletRequest req, HttpServletResponse res) throws ServletException,
			IOException {
		ActionMessages errors = new ActionMessages();
		errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.invalidLogin"));
		req.setAttribute(Globals.ERROR_KEY, errors);
		res.setStatus(HttpServletResponse.SC_FORBIDDEN);
		RequestDispatcher rd = null;
		if (log.isDebugEnabled()) {
			log.debug("Not logged in, redirecting to error page for API logins.");
		}
		rd = req.getRequestDispatcher(apiErrorPath);
		try {
			rd.forward(req, res);
			return;
		} catch (Exception se) {
			return;
		}
	}

	protected void redirectToLoginFormPreservingOriginalRequest(HttpServletRequest req, HttpServletResponse res)
			throws ServletException, IOException {
		RequestDispatcher rd = null;
		if (log.isDebugEnabled()) {
			log.debug("Not logged in, redirecting to login form.");
		}
		// Prevent form submissions from processing
		String origReq = req.getRequestURI();
		origReq = origReq.replaceFirst(req.getContextPath(), "");
		origReq = origReq.replaceAll("ImportPreviewProcess.do", "ImportUploadForm.do");
		origReq = origReq.replaceAll("ImportProcess.do", "ImportUploadForm.do");
		origReq = origReq.replaceAll("Process", "");
		origReq = origReq.replaceAll("QuickAdd.do", ".do");
		origReq = origReq.replaceAll("Add.do", "AddForm.do");
		origReq = origReq.replaceAll("Edit.do", "EditForm.do");
		origReq = origReq.replaceAll("EditDelete.do", ".do");
		origReq = origReq.replaceAll("Logout.do", "Welcome.do");
		origReq = origReq.replaceAll("LuceneAddAll.do", "LuceneForm.do");
		origReq = origReq.replaceAll("ProductSorter.do", "Product.do");
		origReq = origReq.replaceAll("CategorySorter.do", "Category.do");
		origReq = origReq.replaceAll("EmailerInitialize.do", "SettingsEmailer.do");
		if (req.getQueryString() != null) {
			origReq = origReq + "?" + req.getQueryString();
		} else {
			StringBuffer sb = new StringBuffer("");
			Enumeration e = req.getParameterNames();
			while (e.hasMoreElements()) {
				Object nameOb = e.nextElement();
				if (nameOb instanceof String) {
					String name = (String) nameOb;
					if (name.indexOf("ID") > 0 || name.equals("parentName") || name.equals("grandparentName")
							|| name.equals("greatGrandparentName") || name.equals("assignments")
							|| name.equals("editMode") || name.equals("itemsPerPage") || name.equals("sortByString")
							|| name.equals("sortAscOrDesc") || name.equals("firstRow") || name.equals("jumpToSelect")
							|| name.equals("searchString")) {
						sb.append("&" + name + "=");
						try {
							sb.append(URLEncoder.encode(req.getParameter(name), "UTF-8"));
						} catch (UnsupportedEncodingException uee) {
							sb.append(req.getParameter(name));
						}
					}
				}
			}
			origReq = origReq + "?" + sb.toString();
		}
		req.setAttribute("originalRequest", origReq);
		rd = req.getRequestDispatcher(loginFormPath);
		try {
			rd.forward(req, res);
			return;
		} catch (Exception se) {
			return;
		}
	}

	private boolean isLegacyApiRequest(HttpServletRequest request, HttpServletResponse response) {
		RequestDispatcher rd = null;
		if (request.getParameter("userName") != null && request.getParameter("password") != null
				&& !isLoginRequest(request)) {
			// Not logged in but user name and password are present
			if (log.isInfoEnabled()) {
				log.info("Not logged in, user and password are present.");
			}
			String originalRequest = request.getRequestURI();
			originalRequest = originalRequest.replaceFirst(request.getContextPath(), "");
			request.setAttribute("originalRequest", originalRequest);
			rd = request.getRequestDispatcher(loginPath);
			try {
				rd.forward(request, response);
			} catch (Exception se) {
			}
			return false;
		}
		return false;
	}

	/**
	 * Performs custom processing at the end of each call to <code>processPreProcess</code>. Intended to be overridden
	 * by subclasses for customizations. This method should return <code>false</code> to signal
	 * <code>processPreProcess</code> to return <code>false</code> itself.
	 * 
	 * @param request
	 *            The servlet request we are processing
	 * @param response
	 *            The servlet response we are creating
	 * 
	 * @return <code>true</code> if the processing was successful, <code>false</code> if processing should stop.
	 */
	public boolean sscOnAdministratorPreRequestEnd(HttpServletRequest request, HttpServletResponse response) {
		BusinessObjectFactory bof = (BusinessObjectFactory) request.getAttribute("businessObjectFactory");
		if (bof != null) {
			ActionUtils actionUtils = (ActionUtils) bof.createUtilsObject("actionUtilsImplementer");
			return actionUtils.onAdministratorPreRequestEnd(request, response);
		}
		return true;
	}

	/*
	 * Overrides the default Struts permissions check so that a more
	 * user-friendly screen may be presented to the user when they attempt to
	 * access an action they don't have permission for.
	 */
	protected boolean processRoles(HttpServletRequest request, HttpServletResponse response, ActionMapping mapping)
			throws IOException, ServletException {

		String roles[] = mapping.getRoleNames();
		if ((roles == null) || (roles.length < 1)) {
			return (true);
		}

		for (int i = 0; i < roles.length; i++) {
			if (request.isUserInRole(roles[i])) {
				if (log.isDebugEnabled()) {
					log.debug(" User '" + request.getRemoteUser() + "' has role '" + roles[i] + "', granting access");
				}
				return (true);
			}
		}

		if (log.isDebugEnabled()) {
			log.debug(" User '" + request.getRemoteUser() + "' does not have any required role, denying access");
		}

		// ///////////////////////////////////////////////////////////
		forbidRequest(request, response);
		// ///////////////////////////////////////////////////////////

		return (false);

	}
}
