package com.softslate.commerce.businessobjects.core;

import java.util.Collection;
import java.util.Map;
import java.util.Properties;

import com.google.common.eventbus.EventBus;
import com.google.inject.Injector;
import com.softslate.commerce.daos.core.DAOFactory;

/**
 * Interface for the factory class used to create instances of business layer
 * objects,
 * {@link com.softslate.commerce.businessobjects.core.BusinessObjectFactoryImpl}.
 * Objects are created with the <code>createObject</code>,
 * <code>createObjectFromClassName</code>, or <code>createUtilsObject</code>
 * methods.
 * <p>
 * Generally, this class uses its <code>appComponents</code> property to look
 * up the fully-qualified class name of the requested interface (taken from the
 * <code>appComponents.properties</code> file. Then it uses the Java
 * reflection API to create the instance.
 * <p>
 * An instance of <code>BusinessObjectFactory</code> is created in the Struts
 * layer. Its properties are populated there with the application's
 * {@link com.softslate.commerce.daos.core.DAOFactory},
 * {@link com.softslate.commerce.businessobjects.core.Settings},
 * <code>appComponents</code>, <code>appSettings</code>, and with the
 * {@link com.softslate.commerce.businessobjects.core.User} currently accessing
 * the system.
 * <p>
 * Immediately after instantiating the requested business object, this factory
 * passes its own properties to it so that the newly created object can use them
 * to communicate with the rest of the application.
 * 
 * @author David Tobey
 * 
 */
public interface BusinessObjectFactory {

	/**
	 * Retrieves the application settings previously set for this factory. The
	 * application settings are settings for information unique to a given
	 * installation, such as the location of files on the server, as opposed to
	 * the <code>Settings</code>, which stores settings drawn from the
	 * database.
	 * 
	 * @return A <code>Properties</code> object that represents various
	 *         settings for the application.
	 */
	public Properties getAppSettings();

	/**
	 * Sets this factory's application settings. This factory will pass the
	 * application settings to every business object it creates. The application
	 * settings are settings for information unique to a given installation,
	 * such as the location of files on the server, as opposed to the
	 * <code>Settings</code> interface, which stores settings drawn from the
	 * database.
	 * 
	 * @param appSettings
	 *            The current application settings in effect for the system.
	 */
	public void setAppSettings(Properties appSettings);

	/**
	 * Retrieves the application components previously set for this factory. The
	 * components identify which Java class implement the various Interfaces
	 * used by the system.
	 * 
	 * @return A <code>Properties</code> object that represents the components
	 *         for the application.
	 */
	public Properties getAppComponents();

	/**
	 * Sets this factory's application components. This factory will pass the
	 * application components to every business object it creates. The
	 * components identify which Java class implement the various Interfaces
	 * used by the system.
	 * 
	 * @param appComponents
	 *            The current application components in effect for the system.
	 */
	public void setAppComponents(Properties appComponents);

	/**
	 * Retrieves the <code>User</code> previously set for this factory.
	 * 
	 * @return A <code>User</code> object that represents the user currently
	 *         accessing the system.
	 */
	public User getUser();

	/**
	 * Sets this factory's <code>User</code>. This factory will pass the
	 * <code>User</code> to every business object it creates.
	 * 
	 * @param user
	 *            The current <code>User</code> accessing the system, for
	 *            which business processing is to be performed by business
	 *            objects created by this factory.
	 */
	public void setUser(User user);

	/**
	 * Retrieves the <code>Settings</code> previously set for this factory.
	 * 
	 * @return A <code>Settings</code> object that represents various settings
	 *         for the system.
	 */
	public Settings getSettings();

	/**
	 * Sets this factory's <code>Settings</code> instance. This factory will
	 * pass the <code>Settings</code> to every business object it creates.
	 * 
	 * @param settings
	 *            The current <code>Settings</code> for the system, containing
	 *            various settings to be used by business objects created by
	 *            this factory.
	 */
	public void setSettings(Settings settings);

	/**
	 * Retrieves the <code>DAOFactory</code> previously set for this factory.
	 * 
	 * @return A <code>DAOFactory</code> object that can be used to create
	 *         data access objects.
	 */
	public DAOFactory getDaoFactory();

	/**
	 * Sets this factory's <code>DAOFactory</code>. This factory will pass
	 * the <code>DAOFactory</code> to every business object it creates.
	 * 
	 * @param daoFactory
	 *            The current <code>DAOFactory</code> for the system, which
	 *            will allow business objects created by this factory to create
	 *            data access objects.
	 */
	public void setDaoFactory(DAOFactory daoFactory);

	/**
	 * Retrieves the <code>Injector</code> previously set for this factory.
	 * 
	 * @return A <code>Injector</code> object that can be used to create other
	 *         objects and inject them with dependencies.
	 */
	public Injector getInjector();

	/**
	 * Sets this factory's <code>Injector</code>. This factory will pass the
	 * <code>Injector</code> to every business object it creates.
	 * 
	 * @param injector
	 *            The current <code>Injector</code> for the system, which can be
	 *            used to create other objects and inject them with
	 *            dependencies.
	 */
	public void setInjector(Injector injector);

	/**
	 * Retrieves the <code>EventBus</code> previously set for this factory.
	 * 
	 * @return A <code>EventBus</code> object that can be used to post events
	 *         and trigger subscribers to run their code.
	 */
	public EventBus getEventBus();

	/**
	 * Sets this factory's <code>EventBus</code>. This factory will pass the
	 * <code>EventBus</code> to every business object it creates.
	 * 
	 * @param eventBus
	 *            The current <code>EventBus</code> for the system, which can be
	 *            used to post events to subscribers
	 */
	public void setEventBus(EventBus eventBus);

	/**
	 * Creates a business object.
	 * <p>
	 * A lookup is performed first to find the fully-qualified Java class name
	 * of the object to create. The <code>type</code> parameter corresponds to a
	 * key in the <code>appComponents.properties</code> file. This factory uses
	 * its <code>appComponents</code> property to perform the lookup.
	 * <p>
	 * If the Java class name is found, an object is created, and then this
	 * factory populates it with its own properties. These properties allow the
	 * newly create object to communicate with the rest of the application.
	 * 
	 * @param type
	 *            A <code>String</code> associated with the class name of the
	 *            business object to be created.
	 * @return An instance of the requested business object, or
	 *         <code>null</code>, if an error occurs.
	 * @throws Exception
	 */
	public <T extends BusinessObject> T createObject(String type) throws Exception;

	/**
	 * Creates a business object.
	 * <p>
	 * Using the <code>className</code> parameter, this factory creates an
	 * object, and then populates it with its own properties. These properties
	 * allow the newly create object to communicate with the rest of the
	 * application.
	 * 
	 * @param className
	 *            A <code>String</code> associated with the class name of the
	 *            business object to be created.
	 * @return An instance of the requested business object, or
	 *         <code>null</code>, if an error occurs.
	 * @throws Exception
	 */
	public <T extends BusinessObject> T createObjectFromClassName(String className) throws Exception;

	public <T extends BusinessObject> T createObjectFromClassDefinition(Class classDefinition) throws Exception;

	public <T extends BusinessObject> T createObject(Class<T> type);

	/**
	 * Given a type corresponding to a key in the
	 * <code>appComponents.properties</code> file, a method name, and a
	 * <code>Map</code> as the argument, loops through all
	 * <code>BusinessProcessor</code>s configured in
	 * <code>appComponents.properties</code>, calling each in turn. Returns a
	 * <code>Map</code> compling all the results of each
	 * <code>BusinessProcessor</code> called. If any of the
	 * <code>BusinessProcessor</code>s returns a <code>Map</code> with a
	 * key of <code>resultCode</code> that is not equal to "0", processing of
	 * all remaining processors is halted, and this method returns immediately.
	 * 
	 * @param type
	 *            A String such as "activePaymentProcessors" corresponding to a
	 *            key in the <code>appComponents.properties</code> file.
	 * @param method
	 *            The method of each <code>BusinessProcessor</code> to invoke
	 *            in turn.
	 * @param argument
	 *            A <code>Map</code> representing the single argument the
	 *            method accepts.
	 * @return A <code>Map</code> representing all the results of each
	 *         processor combined. They are combined into a single
	 *         <code>Map</code> using the <code>Map.putAll()</code> method.
	 * @throws Exception
	 */
	public Map invokeMultipleProcessors(String type, String method, Map argument)
			throws Exception;

	/**
	 * Given a type corresponding to a key in the
	 * <code>appComponents.properties</code> file, a method name, and a Map as
	 * the argument, loops through all <code>BusinessProcessor</code>s
	 * configured in <code>appComponents.properties</code>, calling each in
	 * turn. Returns a <code>Collection</code> compling all the results of
	 * each <code>BusinessProcessor</code> called.
	 * 
	 * @param type
	 *            A <code>String</code> such as "activeShippingProcessors"
	 *            corresponding to a key in the
	 *            <code>appComponents.properties</code> file.
	 * @param method
	 *            The method of each <code>BusinessProcessor</code> to invoke
	 *            in turn. (E.g. "loadShippingOptions")
	 * @param argument
	 *            A <code>Map</code> representing the single argument the
	 *            method accepts.
	 * @return A <code>Collection</code> representing all the results of each
	 *         processor combined.
	 * @throws Exception
	 */
	public Collection invokeMultipleCollectionProcessors(String type,
			String method, Map argument) throws Exception;

	/**
	 * Given a type corresponding to a key in the
	 * <code>appComponents.properties</code> file, creates and returns a
	 * utility object corresponding to the key.
	 * <p>
	 * If no class can be instantiated for the given class name, an instance of
	 * one of the following classes is automatically returned (depending on the
	 * value of the type parameter):
	 * <ul>
	 * <li>{@link com.softslate.commerce.customer.core.FormUtilsImpl}</li>
	 * <li>{@link com.softslate.commerce.daos.core.DAOUtilsImpl}</li>
	 * <li>{@link BusinessObjectUtilsImpl}</li>
	 * <li>{@link com.softslate.commerce.customer.core.ActionUtilsImpl}</li>
	 * </ul>
	 * 
	 * @param type
	 *            A <code>String</code> such as "formUtilsImplementer"
	 *            corresponding to a key in the
	 *            <code>appComponents.properties</code> file.
	 * @return An instance of the requested business object.
	 */
	public Object createUtilsObject(String type);

	public <T> T createUtilsObject(Class<T> type);

	/**
	 * Creates a <code>Map</code> of this factory's properties. This factory
	 * will pass the <code>Map</code> to the
	 * <code>initialize(Map parameters)</code> method of each business object it
	 * creates, which in turn populates the object's properties.
	 * 
	 * @return A <code>Map</code> of this factory's properties.
	 */
	public Map getTransferMap();

}
