package com.softslate.commerce.businessobjects.order;

import java.util.Collection;
import java.util.Map;

import com.softslate.commerce.businessobjects.core.BusinessProcessor;
import com.softslate.commerce.businessobjects.product.Product;
import com.softslate.commerce.customer.core.BaseForm;

/**
 * Interface for business logic methods having to do with SKUs and inventory.
 * <p>
 * {@link com.softslate.commerce.businessobjects.order.CartProcessor} employs an
 * instance implementing <code>InventoryProcessor</code> when inventory needs
 * to be processed.
 * <p>
 * When the application needs to create an instance that implements
 * <code>InventoryProcessor</code>,
 * {@link com.softslate.commerce.businessobjects.core.BusinessObjectFactory}
 * finds the name of the Java class to instantiate from the
 * "inventoryProcessorImplementer" setting in the
 * <code>appComponents.properties</code> file.
 * <p>
 * The default "inventoryProcessorImplementer" is
 * {@link com.softslate.commerce.businessobjects.order.BasicInventoryProcessor}.
 */
public interface InventoryProcessor extends BusinessProcessor {

	/**
	 * Processes inventory as new <code>OrderItem</code>s are added to a
	 * user's cart. Decrements inventory levels (in the memory objects) if
	 * products are set to be decremented upon being added to a cart. (Objects
	 * are updated in the database by
	 * {@link com.softslate.commerce.businessobjects.order.BasicCartProcessor#processAddItems(Collection)},
	 * which is what calls this method.)
	 * <p>
	 * The default implementation of this method does the following:
	 * <ol>
	 * <li>Loads the inventory settings in effect for each of the products
	 * corresponding to the <code>OrderItem</code>s being added. </li>
	 * <li>Calls
	 * {@link com.softslate.commerce.businessobjects.core.BaseBusinessProcessor#loadMatchingSkus(Collection, Collection, Collection)}
	 * to load the skus corresponding to the <code>OrderItem</code>s. </li>
	 * <li>Calls {@link #processInventory(Collection, Map, Collection, String)}
	 * to process the inventory.</li>
	 * <li>Removes any order items that were rejected because one of their
	 * corresponding SKUs are out of stock.</li>
	 * <li>Adjusts the quantities of any order items whose quantities exceed
	 * the quantity in stock.</li>
	 * <li>Calls {@link CartProcessor#processQuantities(Collection)} to update
	 * the cart with the new quantities.</li>
	 * </ol>
	 * 
	 * @param orderItems
	 *            The <code>Collection</code> of
	 *            {@link com.softslate.commerce.businessobjects.order.OrderItem}
	 *            objects being added to the cart.
	 * @return A <code>Map</code> containing results of the processing. In the
	 *         default implementation, the <code>Map</code> is the same as the
	 *         one returned by
	 *         {@link #processInventory(Collection, Map, Collection, String)}
	 *         (which this method calls).
	 * @throws Exception
	 * @see com.softslate.commerce.businessobjects.order.BasicCartProcessor#processAddItems(Collection)
	 */
	public Map processAddItems(Collection orderItems) throws Exception;

	/**
	 * Checks inventory for a user's cart during checkout. This step takes
	 * places as the user passes through checkout, but before submitting payment
	 * information. Makes sure that if products are to be decremented on
	 * checkout, no skus are out of stock. Inventory does not get decremented
	 * with this method.
	 * <p>
	 * The default implementation of this method largely follows the same steps
	 * as {@link #processAddItems(Collection)}.
	 * 
	 * @param parameters
	 *            The most recent parameters submitted by the user.
	 * @return A <code>Map</code> containing results of the processing. In the
	 *         default implementation, the <code>Map</code> is the same as the
	 *         one returned by
	 *         {@link #processInventory(Collection, Map, Collection, String)}
	 *         (which this method calls).
	 * @throws Exception
	 * @see com.softslate.commerce.customer.order.CheckoutAction#processInventoryDiscountCheck(BaseForm)
	 */
	public Map processOnCheckout(Map parameters) throws Exception;

	/**
	 * Processes inventory for a user's cart as an order is completed. This step
	 * takes place after a successful payment processing. Inventory is
	 * decremented at this point for products set to be decremented on checkout.
	 * 
	 * @param parameters
	 *            The most recent parameters submitted by the user.
	 * @return A <code>Map</code> containing results of the processing. In the
	 *         default implementation, the <code>Map</code> is the same as the
	 *         one returned by
	 *         {@link #processInventory(Collection, Map, Collection, String)}
	 *         (which this method calls).
	 * @throws Exception
	 */
	public Map processOrderComplete(Map parameters) throws Exception;

	/**
	 * Generically processes inventory, returning a Map with information about
	 * items and skus. Does not change the order items being checked.
	 * 
	 * @param orderItems
	 *            A <code>Collection</code> of
	 *            {@link com.softslate.commerce.businessobjects.order.OrderItem}
	 *            objects being processed.
	 * @param productSettings
	 *            A <code>Map</code> of <code>Collections</code>
	 *            representing the inventory settings in effect for the products
	 *            in the <code>OrderItem</code>s.
	 * @param matchingSkus
	 *            A <code>Collection</code> of <code>Map</code>s
	 *            representing all the SKUs associated with the
	 *            <code>OrderItem</code>s.
	 * @param decrementBehavior
	 *            A <code>String</code> representing when inventory is to be
	 *            decremented. Either "onAddingToCart" or "onOrderCompletion".
	 * @return A <code>Map</code> containing results of the processing. In the
	 *         default implementation, the keys of the <code>Map</code>
	 *         include:
	 *         <ul>
	 *         <li><code>skusToDecrement</code>: A <code>Collection</code>
	 *         of <code>Map</code>s, each representing a
	 *         {@link com.softslate.commerce.businessobjects.product.SKU} that
	 *         is to be decremented as result of the processing.</li>
	 *         <li><code>decrementedSKUs</code>: A <code>Map</code> whose
	 *         keys are <code>OrderItem</code> objects and whose values are
	 *         comma-separated <code>String</code>s representing the ids of
	 *         the {@link com.softslate.commerce.businessobjects.product.SKU}
	 *         objects corresponding to the <code>OrderItem</code> that are to
	 *         be decremented. (Used to populate the <code>OrderItem</code>'s
	 *         <code>descrementedSKUs</code> property.)</li>
	 *         <li><code>rejectedOrderItems</code>: A
	 *         <code>Collection</code> of <code>OrderItem</code> objects
	 *         that have been removed from the user's cart due to their being
	 *         out of stock.</li>
	 *         <li><code>adjustedQuantities</code>: A <code>Map</code>
	 *         whose keys are <code>OrderItem</code> objects and whose values
	 *         are <code>String</code>s representing the
	 *         <code>OrderItem</code>s new quantity, after being adjusted
	 *         because not enough inventory is available.</li>
	 *         <li><code>newOOSSKUs</code>: A <code>Collection</code> of
	 *         <code>Map</code>s, each representing a
	 *         {@link com.softslate.commerce.businessobjects.product.SKU} that
	 *         has just reached its out of stock level as result of the
	 *         processing.</li>
	 *         <li><code>productsToDeactivate</code>: A
	 *         <code>Collection</code> of <code>String</code>s, each
	 *         representing the id of a
	 *         {@link com.softslate.commerce.businessobjects.product.Product}
	 *         object that should now be deactivated because a
	 *         {@link com.softslate.commerce.businessobjects.product.SKU}
	 *         associated with it has just gone out of stock.</li>
	 *         <li><code>lowStockEmails</code>: A <code>Collection</code>
	 *         of <code>Map</code>s, each representing a
	 *         {@link com.softslate.commerce.businessobjects.product.SKU} that
	 *         has reached its 'low stock' level, and should trigger a low stock
	 *         email as a result.</li>
	 *         </ul>
	 * @throws Exception
	 */
	public Map processInventory(Collection orderItems, Map productSettings,
			Collection matchingSkus, String decrementBehavior) throws Exception;

	/**
	 * Inspects a Product and returns a Map of useful variables related to the
	 * state of the product's inventory.
	 * 
	 * @param product
	 * @param attributeSKUCollection
	 *            An optional Collection of attribute-only SKUs used to define
	 *            the inventory settings that are returned.
	 * @return A Map of variables related to the product's inventory.
	 */
	public Map defineInventorySettings(Product product,
			Collection attributeSKUCollection);
}
