package com.softslate.commerce.businessobjects.shipping;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import com.softslate.commerce.businessobjects.core.ApplicationException;
import com.softslate.commerce.businessobjects.order.Country;
import com.softslate.commerce.businessobjects.order.OrderDelivery;
import com.softslate.commerce.businessobjects.order.OrderProcessor;
import com.softslate.commerce.businessobjects.order.OrderShippingRule;
import com.softslate.commerce.businessobjects.order.StatesAndCountries;
import com.softslate.commerce.businessobjects.product.Manufacturer;

/**
 * Connects to the USPS server for live online rates. Sends XML rate requests to
 * the USPS server and reads the XML responses from it. From the XML Response,
 * it builds a set of shipping options which get presented to the customer
 * during checkout.
 * 
 * @author ajain, dtobey
 * @see com.softslate.commerce.businessobjects.shipping.ShippingProcessor
 * @see com.softslate.commerce.businessobjects.shipping.BaseShippingProcessor
 * 
 */
public class USPSShippingProcessor extends BaseShippingProcessor implements
		ShippingProcessor {

	public static final String FIRST_CLASS_MAIL_PACKAGE = "First-Class Mail Parcel";

	static Log log = LogFactory.getLog(USPSShippingProcessor.class);

	/*
	 * The total weight of the shipment - all the package weights added
	 * together.
	 */
	public Double totalShipmentWeight = new Double(0);

	/*
	 * USPS Variables, which get populated from sscSettings table
	 */
	public String URL = null;

	public String USERID = null; // <API_NAME @USERID=.. @PASSWORD=..>

	public String SHIPPER_ZIP = null; // <ZipOrigination>...</ZipOrigination>

	public Boolean machinable = new Boolean(false);

	public String uspsExtraCharge = "0";

	public String uspsExtraWeight = "0";

	public String default_Weight_Pounds = null;

	public String default_Weight_Ounces = null;

	public String package_Size = null;

	public boolean USPS_PackageUseDimensions;

	public String default_Width = null;

	public String default_Length = null;

	public String default_Height = null;

	public String USPS_Destination;

	public String uspsDomesticServices;

	public String uspsInternationalServices;

	public boolean domesticService;

	public boolean internationalService;

	public boolean addInsuranceInternational;

	public String default_Girth;

	public StatesAndCountries stateAndCountries;

	public String default_Insurance;

	public double default_Insurance_double = 0.0;

	/*
	 * Load the available options from USPS
	 */
	public Collection loadShippingOptions() throws Exception {
		if (log.isDebugEnabled())
			log.debug("Loading USPS shipping options.");

		setUSPSSettings();

		// Throw an error if a delivery address is not defined.
		if (getUser().getOrder() == null
				|| getUser().getOrder().firstDelivery() == null
				|| getUser().getOrder().firstDelivery().getPostalCode() == null
				|| getUser().getOrder().firstDelivery().getCountry() == null
				|| getUser().getOrder().getOrderShippingRules() == null
				|| getUser().getOrder().getOrderShippingRules().isEmpty()) {
			if (log.isInfoEnabled())
				log.info("Delivery or packages not defined.");
			String errorCode = "ADDRESS";
			String errorMessage = "Delivery or packages not defined.";
			throw new ApplicationException(errorCode + ": " + errorMessage);
		}

		// Identify if it's a domestic or international shipment
		OrderDelivery orderFirstDelivery = getUser().getOrder().firstDelivery();
		String country = orderFirstDelivery.getCountry();
		boolean isDomestic = true;
		if (country.equalsIgnoreCase("US") && this.domesticService) {
			isDomestic = true;
		} else if (!country.equalsIgnoreCase("US") && this.internationalService) {
			isDomestic = false;
		} else {
			if (log.isDebugEnabled())
				log.debug("Domestic/intl option is turned off for this shipment.");
			return new ArrayList();
		}

		// A collection where all the rates from each drop shipper will be added
		// up
		Collection mergedRates = new ArrayList();

		// Find all the drop shippers represented by items in the order, and
		// define the dropShipperIDs variable.
		retrieveDropShippersFromOrder(getUser().getOrder());

		// First query for the shipping rates for everything that does not have
		// a drop shipper, using the global settings.
		if (log.isInfoEnabled())
			log.info("Processing non-drop shipped items.");
		currentDropShipperID = null;
		String data = getXmlFromUsps(isDomestic);
		if (data != null) {
			Map ratesAndErrors = new HashMap();
			if (isDomestic) {
				ratesAndErrors = formatRateV4Rates(data);
			} else {
				ratesAndErrors = formatIntlV2Rates(data);
			}
			Collection rates = (Collection) ratesAndErrors.get("rates");
			String errorCode = (String) ratesAndErrors.get("errorCode");
			String errorMessage = (String) ratesAndErrors.get("errorMessage");
			if ((rates == null || rates.isEmpty()) && errorCode != null
					&& errorMessage != null) {
				if (log.isErrorEnabled())
					log.error(errorCode + ": " + errorMessage);
				throw new ApplicationException(errorCode + ": " + errorMessage);
			}

			// Merge the rates for all drop shippers by adding them up
			mergedRates = mergeRates(mergedRates, rates);
		}

		// Query for each drop shipper now separately, adding the rates together
		// as we go.
		Iterator i = dropShippers.iterator();
		while (i.hasNext()) {
			Manufacturer m = (Manufacturer) i.next();
			SHIPPER_ZIP = m.getPostalCode();
			currentDropShipperID = new Integer(m.getManufacturerID());
			totalShipmentWeight = new Double(0);
			if (log.isInfoEnabled())
				log.info("Processing drop shipper: " + m.getCode()
						+ "; currentDropShipperID: " + currentDropShipperID);
			String dropShipperData = getXmlFromUsps(isDomestic);
			if (dropShipperData != null) {
				Map dropShipperRatesAndErrors = new HashMap();
				if (isDomestic) {
					dropShipperRatesAndErrors = formatRateV4Rates(dropShipperData);
				} else {
					dropShipperRatesAndErrors = formatIntlV2Rates(dropShipperData);
				}

				// Throw an exception if there was an error and no rates
				// returned
				Collection dropShipperRates = (Collection) dropShipperRatesAndErrors
						.get("rates");
				String dropShipperErrorCode = (String) dropShipperRatesAndErrors
						.get("errorCode");
				String dropShipperErrorMessage = (String) dropShipperRatesAndErrors
						.get("errorMessage");
				if ((dropShipperRates == null || dropShipperRates.isEmpty())
						&& dropShipperErrorCode != null
						&& dropShipperErrorMessage != null) {
					throw new ApplicationException(dropShipperErrorCode + ": "
							+ dropShipperErrorMessage);
				}

				// Merge the rates for all drop shippers by adding them up
				mergedRates = mergeRates(mergedRates, dropShipperRates);
			}
		}

		return mergedRates;
	}

	public String getXmlFromUsps(boolean isDomestic) throws Exception {
		if (log.isInfoEnabled())
			log.info("Loading rates from USPS.");

		// Build the XML request
		String queryString = null;
		if (isDomestic) {
			String xmlRequest = createRateV4Request();
			if (xmlRequest == null) {
				return null;
			}
			queryString = "API=RateV4&XML=" + xmlRequest;
		} else {
			String xmlRequest = createIntlRateV2Request();
			if (xmlRequest == null) {
				return null;
			}
			queryString = "API=IntlRateV2&XML=" + xmlRequest;
		}
		if (log.isInfoEnabled())
			log.info("USPS request: " + queryString);

		// Send the request to USPS
		String data = sendHTTPPost(this.URL, queryString, true);
		
		if (log.isInfoEnabled())
			log.info("USPS response: " + data);

		return data;

	}

	public int processShipping() throws Exception {
		return 0;
	}

	public String createRateV4Request() throws Exception {

		boolean foundAPackage = false;

		Document document1 = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		DOMImplementation impl = builder.getDOMImplementation();
		document1 = impl.createDocument(null, "RateV4Request", null);

		Element root = document1.getDocumentElement();
		root.setAttribute("USERID", this.USERID);

		// Add multiple packages from the orderShippingRule table
		if (getUser().getOrder().getOrderShippingRules() != null
				&& !getUser().getOrder().getOrderShippingRules().isEmpty()) {
			int i = 0;
			Iterator j = getUser().getOrder().getOrderShippingRules()
					.iterator();
			while (j.hasNext()) {
				OrderShippingRule osr = (OrderShippingRule) j.next();
				boolean createThePackage = isPackageOkToCreate(osr);
				if (createThePackage) {
					if (log.isInfoEnabled())
						log.info("Processing package: " + osr.getCode());
					Element p1 = createPackageNode(document1, osr);
					if (p1 != null) {
						p1.setAttribute("ID", "" + i++);
						root.appendChild(p1);
						foundAPackage = true;
					}
				}
			}
		}

		// If no packages were found, return null to skip the call to USPS
		if (!foundAPackage) {
			return null;
		}

		StringWriter output1 = new StringWriter();
		try {
			TransformerFactory
					.newInstance()
					.newTransformer()
					.transform(new DOMSource(document1),
							new StreamResult(output1));
		} catch (Exception ex) {
			log.error("Error transforming XML", ex);
		}
		return output1.toString();
	}

	public String createIntlRateV2Request() throws Exception {

		boolean foundAPackage = false;

		Document document1 = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		DOMImplementation impl = builder.getDOMImplementation();
		document1 = impl.createDocument(null, "IntlRateV2Request", null);
		Element root = document1.getDocumentElement();
		root.setAttribute("USERID", this.USERID);
		Element revision = (Element) document1.createElement("Revision");
		revision.appendChild(document1.createTextNode("2"));
		root.appendChild(revision);

		// Now add multiple packages from the orderShippingRule table
		// Note: the origin ZIP is not part of the international API.
		// Nevertheless, when there are multiple drop shippers, separate
		// packages must be sent, which affects the overall rate.
		if (getUser().getOrder().getOrderShippingRules() != null
				&& !getUser().getOrder().getOrderShippingRules().isEmpty()) {
			int i = 0;
			Iterator j = getUser().getOrder().getOrderShippingRules()
					.iterator();
			while (j.hasNext()) {
				OrderShippingRule osr = (OrderShippingRule) j.next();
				boolean createThePackage = isPackageOkToCreate(osr);
				if (createThePackage) {
					if (log.isInfoEnabled())
						log.info("Processing package: " + osr.getCode());

					// Package element
					Element p1 = (Element) document1.createElement("Package");
					p1.setAttribute("ID", "" + i++);

					// Find the weight - which is complicated for the generic
					// package when there are multiple dropshippers
					Double weight = findPackageWeight(osr);
					if (weight == null) {
						continue;
					}

					foundAPackage = true;

					// Add the extra weight
					weight = new Double(weight.doubleValue()
							+ Double.parseDouble(this.uspsExtraWeight));

					// Pounds element
					Element p2 = (Element) document1.createElement("Pounds");
					p2.appendChild(document1.createTextNode(this
							.getWeightInPounds(weight).toString()));
					p1.appendChild(p2);

					// Ounces element
					Element o1 = (Element) document1.createElement("Ounces");
					o1.appendChild(document1.createTextNode(this
							.getWeightInOunces(weight).toString()));
					p1.appendChild(o1);

					// MailType element - always "Package"
					Element mt1 = (Element) document1.createElement("MailType");
					mt1.appendChild(document1.createTextNode("Package"));
					p1.appendChild(mt1);

					// ValueOfContents element
					// Now required
					// if (this.addInsuranceInternational) {
						String valueOfContents = Double
								.toString(this.default_Insurance_double);
						if (null != osr.getSubtotal()
								&& osr.getSubtotal().doubleValue() > this.default_Insurance_double) {
							valueOfContents = osr.getSubtotal().toString();
						}
						Element voc1 = (Element) document1
								.createElement("ValueOfContents");
						voc1.appendChild(document1
								.createTextNode(valueOfContents));
						p1.appendChild(voc1);
					// }

					// Country element
					Element c1 = (Element) document1.createElement("Country");
					String country = this.getCountryFromCode(getUser()
							.getOrder().firstDelivery().getCountry());
					if (country.equals("United Kingdom")) {
						country = "United Kingdom (Great Britain)";
					}
					c1.appendChild(document1.createTextNode(this
							.getCountryFromCode(getUser().getOrder()
									.firstDelivery().getCountry())));
					p1.appendChild(c1);
					boolean didDimensions = doDimensions(document1, p1, osr);
					if (!didDimensions) {

						// IntlV2 needs empty elements at least?
						Element width = (Element) document1.createElement("Width");
						width.appendChild(document1.createTextNode(""));
						p1.appendChild(width);
						Element length = (Element) document1.createElement("Length");
						length.appendChild(document1.createTextNode(""));
						p1.appendChild(length);
						Element height = (Element) document1.createElement("Height");
						height.appendChild(document1.createTextNode(""));
						p1.appendChild(height);
						Element girth = (Element) document1.createElement("Girth");
						girth.appendChild(document1.createTextNode(""));
						p1.appendChild(girth);
					}
					if (country.equals("Canada")) {
						Element oz = (Element) document1.createElement("OriginZip");
						log.info("Shipper ZIP: " + this.SHIPPER_ZIP);
						oz.appendChild(document1.createTextNode(this.SHIPPER_ZIP));
						p1.appendChild(oz);
					}

					// Append this package to the root document
					root.appendChild(p1);
				}
			}
		}

		// If no packages were found, return null to skip the call to USPS
		if (!foundAPackage) {
			return null;
		}

		// Convert to an xml strring
		StringWriter output1 = new StringWriter();

		try {
			TransformerFactory
					.newInstance()
					.newTransformer()
					.transform(new DOMSource(document1),
							new StreamResult(output1));
		} catch (Exception ex) {
			log.error("Error transforming XML", ex);
		}
		return output1.toString();
	}

	public Element createPackageNode(Document document1, OrderShippingRule osr) {
		// Package element
		Element p1 = (Element) document1.createElement("Package");

		// Service element - always "ALL"
		Element s1 = (Element) document1.createElement("Service");
		s1.appendChild(document1.createTextNode("ALL"));
		p1.appendChild(s1);

		// ZipOrigination element
		Element zo1 = (Element) document1.createElement("ZipOrigination");
		zo1.appendChild(document1.createTextNode(this.SHIPPER_ZIP));
		p1.appendChild(zo1);

		// ZipDestination element
		Element zd1 = (Element) document1.createElement("ZipDestination");
		// Get the zip code and format for USPS
		String orderZip = getUser().getOrder().firstDelivery().getPostalCode();
		if (orderZip != null && orderZip.length() > 5) {
			orderZip = orderZip.substring(0, 5);
		}
		zd1.appendChild(document1.createTextNode(orderZip));
		p1.appendChild(zd1);

		// Find the weight - which is complicated for the generic package when
		// there are multiple dropshippers
		Double weight = findPackageWeight(osr);
		if (weight == null) {
			return null;
		}

		// Add the extra weight
		weight = new Double(weight.doubleValue()
				+ Double.parseDouble(this.uspsExtraWeight));

		// Pounds element
		Element p2 = (Element) document1.createElement("Pounds");
		p2.appendChild(document1.createTextNode(this.getWeightInPounds(weight)
				.toString()));
		p1.appendChild(p2);

		// Ounces element
		Element o1 = (Element) document1.createElement("Ounces");
		o1.appendChild(document1.createTextNode(this.getWeightInOunces(weight)
				.toString()));
		p1.appendChild(o1);

		doDimensions(document1, p1, osr);

		// Machinable element
		Element m1 = (Element) document1.createElement("Machinable");
		m1.appendChild(document1.createTextNode(this.machinable.toString()));
		p1.appendChild(m1);
		return p1;

	}

	protected boolean doDimensions(Document document1, Element p1, OrderShippingRule osr) {

		// Add Dimensions if necessary
		// In this case, if it is a [generic] type, we will use the
		// generic settings, otherwise,
		// we will use what comes from the shipping rules table.
		boolean didDimensions = false;
		boolean doDimensions = false;
		if (osr.getPkgExtra2() != null && osr.getPkgExtra2().equals("1")) {
			doDimensions = true;
		} else if (osr.getPkgExtra2() != null && osr.getPkgExtra2().equals("0")) {
			doDimensions = false;
		} else if (this.USPS_PackageUseDimensions) {
			doDimensions = true;
		} else {
			doDimensions = false;
		}

		if (doDimensions) {
			if (log.isDebugEnabled())
				log.debug("Doing Demensions");

			String length = "0";
			String width = "0";
			String height = "0";
			String girth = "0";

			// For generic, we have to use the default package
			// settings
			if (osr.getCode().equals("[generic]")) {
				length = this.default_Length;
				width = this.default_Width;
				height = this.default_Height;
				girth = this.default_Girth;
				if ((girth.equals("") || girth.equals("0"))
						&& !length.equals("") && !width.equals("")
						&& !height.equals("")) {
					girth = doGirth(new Double(length), new Double(width),
							new Double(height)).toString();
				}
			} else {
				length = osr.getPkgLength().toString();
				width = osr.getPkgWidth().toString();
				height = osr.getPkgHeight().toString();
				girth = doGirth(osr.getPkgLength(), osr.getPkgWidth(),
						osr.getPkgHeight()).toString();
			}

			// Skip if all dimensions are 0, or if any are blank.
			if (!length.equals("")
					&& !width.equals("")
					&& !height.equals("")
					&& !girth.equals("")
					&& (!length.equals("0") || !width.equals("0")
							|| !height.equals("0") || !girth.equals("0"))) {

				// Container element
				Element container = (Element) document1.createElement("Container");
				container.appendChild(document1.createTextNode("VARIABLE"));
				p1.appendChild(container);

				// Size element
				String size = doSize(new Double(length), new Double(girth));
				Element s2 = (Element) document1.createElement("Size");
				s2.appendChild(document1.createTextNode(size));
				p1.appendChild(s2);

				// Width element
				Element w1 = (Element) document1.createElement("Width");
				w1.appendChild(document1.createTextNode(width));
				p1.appendChild(w1);

				// Length element
				Element l1 = (Element) document1.createElement("Length");
				l1.appendChild(document1.createTextNode(length));
				p1.appendChild(l1);

				// Height element
				Element h1 = (Element) document1.createElement("Height");
				h1.appendChild(document1.createTextNode(height));
				p1.appendChild(h1);

				// Girth element
				Element g1 = (Element) document1.createElement("Girth");
				g1.appendChild(document1.createTextNode(girth));
				p1.appendChild(g1);
				didDimensions = true;
			}
		}
		if (!didDimensions) {

			// Container element
			Element container = (Element) document1.createElement("Container");
			container.appendChild(document1.createTextNode("VARIABLE"));
			p1.appendChild(container);

			// Size element
			Element s2 = (Element) document1.createElement("Size");
			s2.appendChild(document1.createTextNode(package_Size));
			p1.appendChild(s2);

		}// end of doDimensions

		return didDimensions;
	}

	public Map formatRateV4Rates(String xml) throws Exception {
		Map ratesAndErrors = new HashMap();
		Collection rates = new ArrayList();
		Map ratesMap = new HashMap();
		String currencyCode = "USD";
		boolean stripEstimatedDeliveryDaysFromUSPSMethodName = "1".equals(getSettings().getValue(
				"stripEstimatedDeliveryDaysFromUSPSMethodName"));

		Document doc = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		InputSource is = new InputSource(new StringReader(xml));
		doc = builder.parse(is);

		// Process any errors first
		if (doErrors(doc.getFirstChild(), ratesAndErrors)) {
			return ratesAndErrors;
		}

		// No errors, so process the rates
		Node n = doc.getDocumentElement().getFirstChild();
		while (n != null) {
			if (n.getNodeName().equals("Package")) {
				Node n2 = n.getFirstChild();

				while (n2 != null) {
					if (n2.getNodeName().equals("Error")) {
						doErrors(n2, ratesAndErrors);
					}
					if (n2.getNodeName().equals("Postage")) {
						String code = "";
						String serviceName = "";
						String rate = "";
						code = n2.getAttributes().getNamedItem("CLASSID")
								.getNodeValue();
						Node n3 = n2.getFirstChild();
						while (n3 != null) {

							if (n3.getNodeName().equals("MailService")) {
								serviceName = n3.getFirstChild().getNodeValue();
								serviceName = serviceName.replaceAll("&lt;.*&gt;", "");
								if (stripEstimatedDeliveryDaysFromUSPSMethodName) {
									serviceName = serviceName.replaceAll(" [0-9]-Day", "");
								}
							}
							if (n3.getNodeName().equals("Rate")) {
								rate = n3.getFirstChild().getNodeValue();
							}
							n3 = n3.getNextSibling();
						}

						// Always go with the First-Class Parcel rate - ignore
						// others
						if (!serviceName
								.equalsIgnoreCase(FIRST_CLASS_MAIL_PACKAGE)
								&& code.equals("0")) {
							n2 = n2.getNextSibling();
							continue;
						}
						if (ratesMap.get(code) == null) {
							Map nameAndRate = new HashMap();
							nameAndRate.put("name", serviceName);
							nameAndRate.put("rate", this.uspsExtraCharge);
							ratesMap.put(code, nameAndRate);
						}
						Map nameAndRate = (Map) ratesMap.get(code);
						String name = (String) nameAndRate.get("name");
						String totalRate = (String) nameAndRate.get("rate");
						if (serviceName.equalsIgnoreCase(name)) {
							totalRate = new Double(
									Double.parseDouble(totalRate)
											+ Double.parseDouble(rate))
									.toString();
							nameAndRate.put("rate", totalRate);
						}
					}
					n2 = n2.getNextSibling();
				}
			}
			n = n.getNextSibling();
		}

		// Loop through and create a collection in the right format
		Iterator i = ratesMap.keySet().iterator();
		while (i.hasNext()) {
			String code = (String) i.next();
			if (isInCommaSeparatedString(code, this.uspsDomesticServices)) {
				Map nameAndRate = (Map) ratesMap.get(code);
				String name = (String) nameAndRate.get("name");
				String rate = (String) nameAndRate.get("rate");

				HashMap o = new HashMap();
				o.put("code", code);
				o.put("name", name);
				o.put("priceDouble", new Double(rate));
				o.put("originalPriceDouble", new Double(rate));
				o.put("price",
						formatPrice(Double.parseDouble(rate), currencyCode,
								getUser().getLocale()));
				o.put("originalPrice",
						formatPrice(Double.parseDouble(rate), currencyCode,
								getUser().getLocale()));
				o.put("processorClass", "USPSShippingProcessor");
				o.put("processorName", "USPS");
				o.put("currencyCode", currencyCode);
				Collection discounts = new ArrayList();
				o.put("discounts", discounts);
				rates.add(o);
			}
		}

		ratesAndErrors.put("rates", rates);

		return ratesAndErrors;

	}

	public Map formatIntlV2Rates(String xml) throws Exception {
		Map ratesAndErrors = new HashMap();
		Collection rates = new ArrayList();
		Map ratesMap = new HashMap();
		String currencyCode = "USD";
		boolean stripEstimatedDeliveryDaysFromUSPSMethodName = "1".equals(getSettings().getValue(
				"stripEstimatedDeliveryDaysFromUSPSMethodName"));

		Document doc = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		InputSource is = new InputSource(new StringReader(xml));
		doc = builder.parse(is);

		// Process any errors first
		if (doErrors(doc.getFirstChild(), ratesAndErrors)) {
			return ratesAndErrors;
		}

		// If there were no errors, process the rates
		Node n = doc.getDocumentElement().getFirstChild();
		while (n != null) {
			if (n.getNodeName().equals("Package")) {
				Node n2 = n.getFirstChild();
				while (n2 != null) {
					if (n2.getNodeName().equals("Error")) {
						doErrors(n2, ratesAndErrors);
					}
					if (n2.getNodeName().equals("Service")) {
						String code = "";
						String serviceName = "";
						String rate = "";
						String deliveryDate = "";
						code = n2.getAttributes().getNamedItem("ID")
								.getNodeValue();
						Node n3 = n2.getFirstChild();
						while (n3 != null) {
							if (n3.getNodeName().equals("SvcDescription")) {
								serviceName = n3.getFirstChild().getNodeValue();
								serviceName = serviceName.replaceAll("&lt;.*&gt;", "");
								if (stripEstimatedDeliveryDaysFromUSPSMethodName) {
									serviceName = serviceName.replaceAll(" [0-9]-Day", "");
								}
							}
							if (n3.getNodeName().equals("Postage")) {
								rate = n3.getFirstChild().getNodeValue();
							}
							if (n3.getNodeName().equals("SvcCommitments")) {
								deliveryDate = n3.getFirstChild()
										.getNodeValue();
							}
							n3 = n3.getNextSibling();
						}
						if (ratesMap.get(code) == null) {
							Map nameAndRate = new HashMap();
							nameAndRate.put("name", serviceName);
							nameAndRate.put("rate", this.uspsExtraCharge);
							nameAndRate.put("DeliveryDate", deliveryDate);
							ratesMap.put(code, nameAndRate);
						}
						Map nameAndRate = (Map) ratesMap.get(code);
						String name = (String) nameAndRate.get("name");
						String totalRate = (String) nameAndRate.get("rate");
						if (serviceName.equalsIgnoreCase(name)) {
							totalRate = new Double(
									Double.parseDouble(totalRate)
											+ Double.parseDouble(rate))
									.toString();
							nameAndRate.put("rate", totalRate);
						}
					}
					n2 = n2.getNextSibling();
				}
			}
			n = n.getNextSibling();
		}

		// Loop through and create a collection in the right format
		Iterator i = ratesMap.keySet().iterator();
		while (i.hasNext()) {
			String code = (String) i.next();
			if (isInCommaSeparatedString(code, this.uspsInternationalServices)) {
				Map nameAndRate = (Map) ratesMap.get(code);
				String name = (String) nameAndRate.get("name");
				String rate = (String) nameAndRate.get("rate");
				String deliveryDate = (String) nameAndRate.get("DeliveryDate");

				HashMap o = new HashMap();
				o.put("code", code);
				o.put("name", name);
				o.put("priceDouble", new Double(rate));
				o.put("originalPriceDouble", new Double(rate));
				o.put("price",
						formatPrice(Double.parseDouble(rate), currencyCode,
								getUser().getLocale()));
				o.put("originalPrice",
						formatPrice(Double.parseDouble(rate), currencyCode,
								getUser().getLocale()));
				o.put("processorClass", "USPSShippingProcessor");
				o.put("processorName", "USPS");
				o.put("currencyCode", currencyCode);
				Collection discounts = new ArrayList();
				o.put("discounts", discounts);
				o.put("DeliveryDate", deliveryDate);
				rates.add(o);
			}
		}

		ratesAndErrors.put("rates", rates);

		return ratesAndErrors;

	}

	public boolean doErrors(Node node, Map ratesAndErrors) {
		if (node != null && node.getNodeName().equals("Error")) {
			String errorCode = "Error";
			String errorMessage = "Unknown Error";
			Node n2 = node.getFirstChild();
			while (n2 != null) {
				if (n2.getNodeName().equals("Number")) {
					errorCode = n2.getFirstChild().getNodeValue();
				}
				if (n2.getNodeName().equals("Description")) {
					errorMessage = n2.getFirstChild().getNodeValue();
				}
				n2 = n2.getNextSibling();
			}
			ratesAndErrors.put("errorCode", errorCode);
			ratesAndErrors.put("errorMessage", errorMessage);
			return true;
		}
		return false;
	}

	public void setUSPSSettings() throws Exception {
		this.SHIPPER_ZIP = (String) getSettings().getValue("uspsZIPOrigin");
		this.USERID = (String) getSettings().getValue("uspsUser");
		this.URL = (String) getSettings().getValue("uspsURL");
		this.uspsExtraCharge = (String) getSettings().getValue(
				"uspsExtraCharge");
		this.uspsExtraWeight = (String) getSettings().getValue(
				"uspsExtraWeight");

		// Domestic, International
		this.USPS_Destination = (String) getSettings().getValue(
				"USPS_Destination");
		this.domesticService = this.isInCommaSeparatedString("01",
				this.USPS_Destination);
		this.internationalService = this.isInCommaSeparatedString("02",
				this.USPS_Destination);

		// IDs of Priority Flat, express, parcel etc..
		this.uspsDomesticServices = (String) getSettings().getValue(
				"uspsDomesticServices");
		this.uspsInternationalServices = (String) getSettings().getValue(
				"uspsIntlServices");

		this.machinable = new Boolean(getSettings().getValue("uspsMachinable")
				.equals("1"));
		this.package_Size = (String) getSettings()
				.getValue("uspsContainerSize");
		this.USPS_PackageUseDimensions = getSettings().getValue(
				"USPS_PackageUseDimensions").equals("1");
		this.default_Length = (String) getSettings().getValue(
				"USPS_PackageDimensionsLength");
		this.default_Height = (String) getSettings().getValue(
				"USPS_PackageDimensionsHeight");
		this.default_Width = (String) getSettings().getValue(
				"USPS_PackageDimensionsWidth");
		this.default_Girth = (String) getSettings().getValue(
				"USPS_PackageDimensionsGirth");

		this.addInsuranceInternational = getSettings().getValue(
				"USPS_PackageUseInsurance").equals("1");
		this.default_Insurance = (String) getSettings().getValue(
				"USPS_PackageMinimumInsuranceValue");

		if (null != this.default_Insurance)
			this.default_Insurance_double = new Double(this.default_Insurance)
					.doubleValue();

		OrderProcessor orderProcessor = (OrderProcessor) this
				.getBusinessObjectFactory().createObject(
						"orderProcessorImplementer");
		this.stateAndCountries = orderProcessor.loadStatesAndCountries();
	}

	public String getCountryFromCode(String country) {
		return ((Country) this.stateAndCountries.getCountryMap().get(country))
				.getName();
	}

	public Integer getWeightInOunces(Double weight) {
		double ounces = Math.ceil((weight.doubleValue() - Math.floor(weight
				.doubleValue())) * 100);
		return new Integer(new Double(ounces * 16 / 100).intValue());
	}

	public Integer getWeightInPounds(Double weight) {
		return new Integer(weight.intValue());
	}

	public String doSize(Double length, Double girth) {
		if (length.doubleValue() + girth.doubleValue() > 84) {
			if (length.doubleValue() + girth.doubleValue() > 108) {
				return "OVERSIZE";
			} else {
				return "LARGE";
			}
		} else {
			return "REGULAR";
		}
	}

	public Double doGirth(Double length, Double width, Double height) {
		if (length == null || width == null || height == null) {
			return null;
		}
		double smallDimension1 = 0;
		double smallDimension2 = 0;
		if (length.doubleValue() >= width.doubleValue()
				&& length.doubleValue() >= height.doubleValue()) {
			smallDimension1 = width.doubleValue();
			smallDimension2 = height.doubleValue();
		} else if (width.doubleValue() > length.doubleValue()
				&& width.doubleValue() >= height.doubleValue()) {
			smallDimension1 = length.doubleValue();
			smallDimension2 = height.doubleValue();
		} else if (height.doubleValue() > length.doubleValue()
				&& height.doubleValue() > width.doubleValue()) {
			smallDimension1 = width.doubleValue();
			smallDimension2 = length.doubleValue();
		}
		double girth = 2 * (smallDimension1 + smallDimension2);
		return new Double(girth);
	}

	public boolean isInCommaSeparatedString(String value,
			String commaSeparatedString) {

		if (null == value || null == commaSeparatedString)
			return false;

		// For USPS, the single-digit services are stored with a 0 in front, so
		// add it to the value before comparing
		if (value.length() == 1) {
			value = "0" + value;
		}

		boolean result = false;
		String[] separatedValues = commaSeparatedString.split(",");
		for (int j = 0; j < separatedValues.length; j++) {
			String thisValue = separatedValues[j].trim();
			if (value.equals(thisValue)) {
				result = true;
				break;
			}
		}
		return result;
	}
}
