package com.softslate.commerce.customer.core;

import java.io.File;
import java.io.FileOutputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import javax.servlet.GenericServlet;
import javax.servlet.ServletException;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.beanutils.SuppressPropertiesBeanIntrospector;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.action.PlugIn;
import org.apache.struts.config.ModuleConfig;

import com.softslate.commerce.daos.core.DAOFactory;
import com.softslate.commerce.daos.core.LiquibaseDAO;

/**
 * Initializes singleton objects in the application scope. Called by Struts as a <code>PlugIn</code> as the application
 * is loaded, and by {@link com.softslate.commerce.customer.core.BaseRequestProcessor} at the beginning of a request, if
 * the singletons are found to be missing.
 * 
 * @author David Tobey
 * 
 */
public class ConfigPlugIn implements PlugIn {

	static Log log = LogFactory.getLog(ConfigPlugIn.class);

	public ConfigPlugIn() {
	}

	public ConfigPlugIn(ActionServlet servlet) {
		setServlet(servlet);
	}

	public ConfigPlugIn(GenericServlet servlet) {
		setServlet(servlet);
	}

	private GenericServlet servlet = null;

	public GenericServlet getServlet() {
		return (this.servlet);
	}

	public void setServlet(GenericServlet servlet) {
		this.servlet = servlet;
	}

	public void init(ActionServlet servlet, ModuleConfig moduleConfig)
			throws ServletException {
		if (log.isInfoEnabled()) {
			log.info("Initializing application configurations.");
		}
		setServlet(servlet);
		initPropertyUtilsConfig();
		initializeAppSettings();
		initializeAppComponents();
		processPreDatabaseUpgrades();
		initializeDAOFactory();
		processLiquibaseUpdates();
		processPostDatabaseUpgrades();
		return;
	}

	/*
	 * Addresses Struts vulnerabilities: CVE-2014-0094, CVE-2014-0114, CVE-2016-1181, CVE-2016-1182
	 */
	protected void initPropertyUtilsConfig() {
		HashSet suppressProperties = new HashSet();
		suppressProperties.add("class");
		suppressProperties.add("multipartRequestHandler");
		suppressProperties.add("resultValueMap");
		PropertyUtils.addBeanIntrospector(new SuppressPropertiesBeanIntrospector(suppressProperties));
		PropertyUtils.clearDescriptors();
	}

	/**
	 * Creates an instance of <code>java.util.Properties</code> representing
	 * application settings in the application scope. Reads in the properties
	 * from the "/WEB-INF/classes/appSettings.properties" file, using
	 * {@link com.softslate.commerce.customer.core.AppUtils#loadProperties}.
	 */
	public void initializeAppSettings() {
		synchronized (getServlet().getServletContext()) {
			String appSettingsPropertiesFile = "appSettings.properties";
			if (System.getProperty("appSettings.configuration") != null) {
				appSettingsPropertiesFile = System
						.getProperty("appSettings.configuration");
			}
			Properties settings = AppUtils
					.loadProperties(appSettingsPropertiesFile, getServlet()
							.getServletContext());
			getServlet().getServletContext().setAttribute("appSettings",
					settings);
			if (log.isErrorEnabled() && settings == null) {
				log.error("Cannot initialize application settings (appSettings.properties).");
			}
		}
	}

	/**
	 * Creates an instance of <code>java.util.Properties</code> representing
	 * component settings in the application scope. Component settings define
	 * which classes implement the various application interfaces used by the
	 * application's code. Reads in the properties from the
	 * "/WEB-INF/classes/appComponents.properties" file, using
	 * {@link com.softslate.commerce.customer.core.AppUtils#loadProperties}.
	 */
	public void initializeAppComponents() throws ServletException {
		synchronized (getServlet().getServletContext()) {
			Properties defaults = null;
			Properties appComponentsDefaults = AppUtils.loadProperties(
					"/appComponents.properties", defaults);
			log.info("appComponentsDefaults: " + appComponentsDefaults);
			Properties settings = AppUtils.loadProperties(
					"/appComponents-custom.properties", appComponentsDefaults);
			if (settings == null) {
				String errorMessage = "appComponents-custom.properties does not "
						+ "exist! Before you upgrade to version 3.0 you "
						+ "must rename /WEB-INF/classes/appComponents.properties "
						+ "to appComponents-custom.properties. Please restore"
						+ "appComponents.properties from a backup and rename it "
						+ "to appComponents-custom.properties before proceeding with the upgrade.";
				if (log.isErrorEnabled())
					log.error(errorMessage);
				throw new ServletException(errorMessage);
			}
			getServlet().getServletContext().setAttribute("appComponents",
					settings);
			log.info("appComponents: " + settings);
			getServlet().getServletContext().setAttribute(
					"appComponentsDefaults", appComponentsDefaults);
		}
	}

	/**
	 * Creates a configured instance of <code>DAOFactory</code> in the
	 * application scope.
	 * <p>
	 * If successful, the "daoFactory" attribute of the servlet context will
	 * contain a <code>DAOFactory</code> that allows the DAOs it creates to to
	 * communicate with the database through Hibernate.
	 */
	public void initializeDAOFactory() {
		synchronized (getServlet().getServletContext()) {
			DAOFactory daoFactory = null;
			// Explicitly load the hibernate.properties file in case the
			// installer was just run.
			String hibernatePropertiesFile = "hibernate.properties";
			if (System.getProperty("hibernate.configuration") != null) {
				hibernatePropertiesFile = System
						.getProperty("hibernate.configuration");
			}
			Properties hibernateSettings = AppUtils.loadProperties(
					hibernatePropertiesFile, getServlet().getServletContext());
			String connectionURL = hibernateSettings
					.getProperty("hibernate.connection.url");
			if (connectionURL == null || connectionURL.endsWith("databasename")) {
				log.error("Connection URL in hibernate properties file ends with 'databasename'"
						+ " - app does not appear to be installed");
				return;
			}
			Properties appComponents = (Properties) getServlet()
					.getServletContext().getAttribute("appComponents");
			Properties appComponentsDefaults = (Properties) getServlet()
					.getServletContext().getAttribute("appComponentsDefaults");
			Properties appSettings = (Properties) getServlet()
					.getServletContext().getAttribute("appSettings");
			String className = appComponents
					.getProperty("daoFactoryImplementer");
			try {
				Class classDefinition = Class.forName(className);
				daoFactory = (DAOFactory) classDefinition.newInstance();
			} catch (Exception e) {
				log.error("Error creating DAOFactory of type " + className, e);
			}
			try {
				daoFactory.setHibernateSettings(hibernateSettings);
				daoFactory.setAppComponents(appComponents);
				daoFactory.setAppComponentsDefaults(appComponentsDefaults);
				daoFactory.setAppSettings(appSettings);
				daoFactory.initialize();
				getServlet().getServletContext().setAttribute("daoFactory",
						daoFactory);
			} catch (Exception e) {
				log.error("Couldn't initialize DAOFactory", e);
			}
		}
		return;
	}

	protected void processLiquibaseUpdates() {
		DAOFactory daoFactory = (DAOFactory) getServlet().getServletContext().getAttribute("daoFactory");
		if (daoFactory != null) {

			// If this is the Community Edition, catch the ClassNotFoundException and just return.
			LiquibaseDAO liquibaseDAO = null;
			try {
				liquibaseDAO = daoFactory.createDAO(LiquibaseDAO.class);
			} catch (RuntimeException re) {
				if (re.getCause() != null) {
					try {
						throw re.getCause();
					} catch (ClassNotFoundException cnfe) {
						log.warn("Cannot find implementer for LiquibaseDAO (swallowing). "
								+ "This is exception is normal in the Community Edition.");
						return;
					} catch (Throwable e) {
						// Swallow here
					}
				}
				throw re;
			}

			// Run the updates, if any.
			liquibaseDAO.initializeLiquibaseAndRunUpdates(getServlet().getServletContext());
		}
	}

	/**
	 * Close the Hibernate SessionFactory as the application shuts down.
	 */
	public void destroy() {
		if (getServlet() == null) {
			return;
		}
		DAOFactory daoFactory = (DAOFactory) getServlet().getServletContext()
				.getAttribute("daoFactory");
		if (daoFactory == null) {
			return;
		}
		if (daoFactory.getSessionFactory() == null) {
			return;
		}
		if (!daoFactory.getSessionFactory().isClosed()) {
			if (log.isInfoEnabled())
				log.info("Closing daoFactory.getSessionFactory()");
			daoFactory.getSessionFactory().close();
		}
	}

	public void processPreDatabaseUpgrades() {
		processUpgrade2dot1();
		processUpgrade2dot2();
		processUpgrade2dot2dot5();
		processUpgrade2dot3dot5();
	}

	public void processPostDatabaseUpgrades() {
		processUpgrade3dot0dot0();
	}

	public void processUpgrade2dot1() {
		// Add default configurations for Manufacturers if they don't exist, so
		// upgrades can proceed.
		boolean componentsAdded = false;
		Properties appComponents = (Properties) getServlet()
				.getServletContext().getAttribute("appComponents");
		if (appComponents.getProperty("manufacturerImplementer") == null) {
			appComponents
					.setProperty("manufacturerImplementer",
							"com.softslate.commerce.businessobjects.product.ManufacturerBean");
			componentsAdded = true;
		}
		if (appComponents.getProperty("manufacturerProcessorImplementer") == null) {
			appComponents
					.setProperty("manufacturerProcessorImplementer",
							"com.softslate.commerce.businessobjects.product.BasicManufacturerProcessor");
			componentsAdded = true;
		}
		if (appComponents.getProperty("manufacturerGatewayDAOImplementer") == null) {
			appComponents
					.setProperty("manufacturerGatewayDAOImplementer",
							"com.softslate.commerce.daos.product.ManufacturerGatewayDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("manufacturerDAOImplementer") == null) {
			appComponents
					.setProperty("manufacturerDAOImplementer",
							"com.softslate.commerce.daos.product.ManufacturerDAOHibernate");
			componentsAdded = true;
		}
		if (componentsAdded) {
			synchronized (getServlet().getServletContext()) {
				try {
					saveComponents(appComponents);
					getServlet().getServletContext().setAttribute(
							"appComponents", appComponents);
					if (log.isWarnEnabled())
						log.warn("Saved new components in appComponents.properties: manufacturerImplementer, manufacturerProcessorImplementer, manufacturerGatewayDAOImplementer, manufacturerDAOImplementer");
				} catch (Exception e) {
					log.error("Couldn't save components", e);
				}
			}
		}
		return;
	}

	public void processUpgrade2dot2() {
		// Add default configurations for Manufacturers if they don't exist, so
		// upgrades can proceed.
		boolean componentsAdded = false;
		Properties appComponents = (Properties) getServlet()
				.getServletContext().getAttribute("appComponents");
		if (appComponents.getProperty("shippingRuleImplementer") == null) {
			appComponents
					.setProperty("shippingRuleImplementer",
							"com.softslate.commerce.businessobjects.shipping.ShippingRuleBean");
			componentsAdded = true;
		}
		if (appComponents.getProperty("skuShippingRuleImplementer") == null) {
			appComponents
					.setProperty("skuShippingRuleImplementer",
							"com.softslate.commerce.businessobjects.shipping.SKUShippingRuleBean");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleRangeImplementer") == null) {
			appComponents
					.setProperty("shippingRuleRangeImplementer",
							"com.softslate.commerce.businessobjects.shipping.ShippingRuleRangeBean");
			componentsAdded = true;
		}
		if (appComponents.getProperty("orderShippingRuleImplementer") == null) {
			appComponents
					.setProperty("orderShippingRuleImplementer",
							"com.softslate.commerce.businessobjects.order.OrderShippingRuleBean");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleGatewayDAOImplementer") == null) {
			appComponents
					.setProperty("shippingRuleGatewayDAOImplementer",
							"com.softslate.commerce.daos.shipping.ShippingRuleGatewayDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleDAOImplementer") == null) {
			appComponents
					.setProperty("shippingRuleDAOImplementer",
							"com.softslate.commerce.daos.shipping.ShippingRuleDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleProcessorImplementer") == null) {
			appComponents
					.setProperty("shippingRuleProcessorImplementer",
							"com.softslate.commerce.businessobjects.shipping.BasicShippingRuleProcessor");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleRangeGatewayDAOImplementer") == null) {
			appComponents
					.setProperty("shippingRuleRangeGatewayDAOImplementer",
							"com.softslate.commerce.daos.shipping.ShippingRuleRangeGatewayDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleRangeDAOImplementer") == null) {
			appComponents
					.setProperty("shippingRuleRangeDAOImplementer",
							"com.softslate.commerce.daos.shipping.ShippingRuleRangeDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("shippingRuleRangeProcessorImplementer") == null) {
			appComponents
					.setProperty(
							"shippingRuleRangeProcessorImplementer",
							"com.softslate.commerce.businessobjects.shipping.BasicShippingRuleRangeProcessor");
			componentsAdded = true;
		}
		if (appComponents.getProperty("orderShippingRuleGatewayDAOImplementer") == null) {
			appComponents
					.setProperty("orderShippingRuleGatewayDAOImplementer",
							"com.softslate.commerce.daos.order.OrderShippingRuleGatewayDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("orderShippingRuleDAOImplementer") == null) {
			appComponents
					.setProperty("orderShippingRuleDAOImplementer",
							"com.softslate.commerce.daos.order.OrderShippingRuleDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("importDAOImplementer") == null) {
			appComponents
					.setProperty("importDAOImplementer",
							"com.softslate.commerce.daos.importexport.ImportDAOHibernate");
			componentsAdded = true;
		}
		if (appComponents.getProperty("importProcessorImplementer") == null) {
			appComponents
					.setProperty("importProcessorImplementer",
							"com.softslate.commerce.businessobjects.importexport.BasicImportProcessor");
			componentsAdded = true;
		}
		if (appComponents.getProperty("exportProcessorImplementer") == null) {
			appComponents
					.setProperty("exportProcessorImplementer",
							"com.softslate.commerce.businessobjects.importexport.BasicExportProcessor");
			componentsAdded = true;
		}
		if (appComponents.getProperty("ordersExportImplementer") == null) {
			appComponents
					.setProperty("ordersExportImplementer",
							"com.softslate.commerce.businessobjects.importexport.OrdersExportBean");
			componentsAdded = true;
		}
		if (componentsAdded) {
			synchronized (getServlet().getServletContext()) {
				try {
					saveComponents(appComponents);
					getServlet().getServletContext().setAttribute(
							"appComponents", appComponents);
					if (log.isWarnEnabled())
						log.warn("Saved new components in appComponents.properties for 2.2 upgrade.");
				} catch (Exception e) {
					if (log.isWarnEnabled())
						log.warn("Couldn't save components: " + e.toString());
				}
			}
		}
		return;
	}

	public void processUpgrade2dot2dot5() {
		// Add default configurations for Lucene advanced search
		boolean componentsAdded = false;
		Properties appComponents = (Properties) getServlet()
				.getServletContext().getAttribute("appComponents");
		if (appComponents.getProperty("searchListImplementer") == null) {
			appComponents
					.setProperty("searchListImplementer",
							"com.softslate.commerce.businessobjects.product.SearchListBean");
			componentsAdded = true;
		}
		if (componentsAdded) {
			synchronized (getServlet().getServletContext()) {
				try {
					saveComponents(appComponents);
					getServlet().getServletContext().setAttribute(
							"appComponents", appComponents);
					if (log.isWarnEnabled())
						log.warn("Saved new components in appComponents.properties: searchListImplementer");
				} catch (Exception e) {
					log.error("Couldn't save components", e);
				}
			}
		}
		return;
	}

	public void processUpgrade2dot3dot5() {
		// Add default configurations for Lucene advanced search
		boolean componentsAdded = false;
		Properties appComponents = (Properties) getServlet()
				.getServletContext().getAttribute("appComponents");
		if (appComponents.getProperty("luceneIndexerImplementer") == null) {
			appComponents
					.setProperty("luceneIndexerImplementer",
							"com.softslate.commerce.businessobjects.product.BasicLuceneIndexer");
			componentsAdded = true;
		}
		if (appComponents.getProperty("luceneSearcherImplementer") == null) {
			appComponents
					.setProperty("luceneSearcherImplementer",
							"com.softslate.commerce.businessobjects.product.BasicLuceneSearcher");
			componentsAdded = true;
		}
		if (componentsAdded) {
			synchronized (getServlet().getServletContext()) {
				try {
					saveComponents(appComponents);
					getServlet().getServletContext().setAttribute(
							"appComponents", appComponents);
					if (log.isWarnEnabled())
						log.warn("Saved new components in appComponents.properties: luceneIndexerImplementer, luceneSearcherImplementer");
				} catch (Exception e) {
					log.error("Couldn't save components", e);
				}
			}
		}
		return;
	}

	public void processUpgrade3dot0dot0() {

		// This method eliminates duplicate properties from
		// appComponents-custom.properties that already exist in the default
		// appComponent.properties file
		// Load the new custom components file
		Properties customProperties = AppUtils.loadProperties(
				"/appComponents-custom.properties", (Properties) null);

		// Load the default components file
		Properties defaultProperties = AppUtils.loadProperties(
				"/appComponents.properties", (Properties) null);

		// New set of cleaned-up custom components
		Properties customComponents = new Properties();

		// Loop through the custom properties so we can save only those that
		// differ from the defaults
		boolean resaveCustomComponents = false;
		Set keys = customProperties.keySet();
		Iterator i = keys.iterator();
		while (i.hasNext()) {
			String key = (String) i.next();
			String value = (String) customProperties.getProperty(key);

			// Skip if the custom value is the same as the default
			if (value.equals(defaultProperties.getProperty(key))) {
				if (log.isDebugEnabled())
					log.debug("Skipping component - same as the default: "
							+ key);
				resaveCustomComponents = true;
				continue;
			}
			customComponents.setProperty(key, value);
		}

		// Save the cleaned-up custom components
		if (resaveCustomComponents) {
			try {
				saveComponents(customComponents);
				if (log.isWarnEnabled())
					log.warn("Saved custom components to appComponents-custom.properties");
			} catch (Exception e) {
				log.error("Couldn't save components", e);
			}
		}
	}

	/**
	 * Saves a new set of properties to the
	 * /WEB-INF/classes/appComponents.properties file.
	 * 
	 * @param appComponents
	 *            The new set of properties to be saved.
	 * @throws Exception
	 */
	public void saveComponents(Properties appComponents) throws Exception {
		String appComponentsPath = getServlet()
				.getServletContext()
				.getRealPath("/WEB-INF/classes/appComponents-custom.properties");
		File appComponentsFile = new File(appComponentsPath);
		FileOutputStream fos = new FileOutputStream(appComponentsFile);
		if (log.isDebugEnabled())
			log.debug("Attempting to write components to " + appComponentsFile
					+ " appComponents: " + appComponents);
		appComponents
				.store(fos,
						"Generated by com.softslate.commerce.customer.core.ConfigPlugin.");
	}

}
