package com.softslate.commerce.customer.customer;

import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.customer.CustomerProcessor;
import com.softslate.commerce.customer.core.BaseForm;
import com.softslate.commerce.customer.order.CartAddAction;

/**
 * Struts action class used for requests to pick up a saved cart by a customer
 * and make it the active cart for the session. Used for requests to
 * "/SavedCartPickup.do".
 * <p>
 * The <code>execute</code> method of this class invokes
 * {@link com.softslate.commerce.businessobjects.customer.CustomerProcessor#loadSavedCart(Map)}
 * to load the details of a given saved cart into the request. If the order is
 * loaded successfully, the instance of
 * {@link com.softslate.commerce.businessobjects.order.Order} is set as the
 * session's current cart by using <code>getUser().setOrder(Order order)</code>.
 * If the order cannot be found in the database, the <code>ActionForward</code>
 * named "failure" will be returned.
 * 
 * @author David Tobey
 * 
 */
public class SavedCartPickupAction extends CartAddAction {
	static Log log = LogFactory.getLog(SavedCartPickupAction.class);

	public ActionForward execute(ActionMapping mapping, ActionForm form,
			HttpServletRequest request, HttpServletResponse response)
			throws Exception {
		if (log.isInfoEnabled())
			log.info("Starting action.");
		BaseForm baseForm = (BaseForm) form;
		synchronized (baseForm.getRequest().getSession()) {

			// Preserve the original request so the customer is redirected
			// back to the main saved carts page after logging in
			request.setAttribute("originalRequest", "/AccountSavedCarts.do");

			// If the user is not logged in, it's an error.
			if (baseForm.getUser().getCustomer() == null) {
				baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE,
						new ActionMessage("errors.userNotLoggedIn"));
				saveErrors(baseForm.getRequest(), (ActionMessages) baseForm
						.getErrors());
				if (log.isInfoEnabled() && !baseForm.getErrors().isEmpty())
					log.info("Validation error(s): "
							+ baseForm.getErrors().toString());
				return mapping.findForward("notLoggedIn");
			}

			// Check the global allowSavedCarts flag
			if (!baseForm.getSettings().getValue("allowSavedCarts").equals("1")) {
				baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.cartsMayNotBeSaved"));
				saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
				if (log.isInfoEnabled() && !baseForm.getErrors().isEmpty())
					log.info("Validation error(s): " + baseForm.getErrors().toString());
				return mapping.findForward("failure");
			}

			// Load the saved cart. This will either duplicate the saved order
			// or reprocess the saved order's items. In either case,
			// getUser().getOrder() will be populated.
			CustomerProcessor customerProcessor = (CustomerProcessor) baseForm
					.getBusinessObjectFactory().createObject(
							"customerProcessorImplementer");
			Map results = customerProcessor.loadSavedCart(PropertyUtils
					.describe(baseForm));

			// If the order is null, it means it is too old
			if (baseForm.getUser().getOrder() == null) {
				baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE,
						new ActionMessage("errors.cartMayNotBePickedUp"));
				saveErrors(baseForm.getRequest(), (ActionMessages) baseForm
						.getErrors());
				if (log.isInfoEnabled() && !baseForm.getErrors().isEmpty())
					log.info("Validation error(s): "
							+ baseForm.getErrors().toString());
				return mapping.findForward("failure");
			}

			// Set the shippingAndTaxesProcessed flag based on if the order's
			// tax is null or not
			if (baseForm.getUser().getOrder().getTax() != null
					&& baseForm.getUser().getOrder().getTaxableSubtotal() != null) {
				baseForm.getRequest().getSession().setAttribute(
						"shippingAndTaxesProcessed", "true");
				baseForm.getRequest().getSession().setAttribute(
						"validatedPaymentForm", null);
				baseForm.getRequest().getSession().setAttribute(
						"processorName", null);
			}
			baseForm.getMessages().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("messages.cartPickedUp"));
			saveMessages(baseForm.getRequest(), baseForm.getMessages());

			// Process inventory actions (if order items were reprocessed).
			if (results != null) {
				String forward = processResults(results, baseForm);
				return mapping.findForward(forward);
			}

			return mapping.findForward("success");
		}
	}
}
