package com.softslate.commerce.customer.product;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.product.Category;
import com.softslate.commerce.businessobjects.product.Product;
import com.softslate.commerce.businessobjects.product.ProductProcessor;
import com.softslate.commerce.businessobjects.product.ProductReview;
import com.softslate.commerce.businessobjects.product.ProductSetting;
import com.softslate.commerce.customer.core.BaseAction;
import com.softslate.commerce.customer.core.BaseDynaForm;

/**
 * Struts action class that corresponds to the product page. Used for requests
 * to "/Product.do".
 * <p>
 * The <code>execute</code> method of this class populates a product's
 * information from the database and puts it in the request scope for use by the
 * view layer. Given a
 * {@link com.softslate.commerce.customer.product.ProductForm} with its
 * <code>code</code> property populated, it creates and initializes an instance
 * of {@link com.softslate.commerce.businessobjects.product.Product}. It then
 * invokes
 * {@link com.softslate.commerce.businessobjects.product.ProductProcessor#getProductFromCode(Product)}
 * to load the product's information from the database. It puts the initialized
 * <code>Product</code> object in the request scope as an attribute with the
 * names "productForm" and "product". A <code>Map</code> representing the
 * product's custom settings and other settings is also placed in the request
 * scope as an attribute named "productSettings".
 * <p>
 * The <code>execute</code> method next loads a <code>Map</code> of
 * attribute-only SKUs and places it in the request as an attribute named
 * "attributeSKUs" so they may be displayed on the product page. (Attribute-only
 * SKUs are SKUs which are not tied to any particular product but which may
 * still be relevant to the product being displayed.)
 * <p>
 * Next, the <code>execute</code> method determines the user's "current
 * category" for purposes of breadcrumbing and places it in a session attribute
 * named "currentCategory", holding the <code>code</code> property of the user's
 * current category. If the user has clicked a link from a category page to
 * arrive at the product page, a request parameter named "category" is used.
 * Otherwise, for example if the user has arrived from the search results page
 * or some other source, the product's <code>primaryCategory</code> is used. The
 * current category determines what category appears above the product in the
 * breadcrumbing links, as well as what category is highlighted in the category
 * tree.
 * <p>
 * Finally, three other request attributes are set from the product's data. Each
 * of these variables is initialized with default values in
 * {@link com.softslate.commerce.customer.core.BaseRequestProcessor#processPreprocess(HttpServletRequest, HttpServletResponse)}
 * and altered here with product-specific data:
 * <ul>
 * <li><code>htmlHeadTitle</code>: The product's <code>name</code> is appended
 * to this value (which is initialized with the <code>storeName</code> setting
 * from the <code>sscSetting</code> database table). By default this variable is
 * used in the &lt;title&gt; tag of the html header.</li>
 * <li><code>htmlHeadKeywords</code>: This value is replaced with the product's
 * <code>keywords</code> property. By default this variable is used in the
 * &lt;meta&gt;/keywords tag of the html header.</li>
 * <li><code>htmlHeadDescription</code>: This value is replaced with the
 * product's <code>shortDescription</code> property. By default this variable is
 * used in the &lt;meta&gt;/description tag of the html header.</li>
 * </ul>
 * 
 * @author David Tobey
 * 
 * @see com.softslate.commerce.customer.product.ProductForm
 * @see com.softslate.commerce.businessobjects.product.Product
 * @see com.softslate.commerce.businessobjects.product.ProductProcessor
 */
public class ProductAction extends BaseAction {

	static Log log = LogFactory.getLog(ProductAction.class);

	public ActionForward execute(ActionMapping mapping, ActionForm form,
			HttpServletRequest request, HttpServletResponse response)
			throws Exception {
		if (log.isInfoEnabled())
			log.info("Starting action.");
		BaseDynaForm baseForm = (BaseDynaForm) form;

		// Load the product
		String useSEOCodeForProductURLs = (String) baseForm.getSettings()
				.getValue("useSEOCodeForProductURLs");
		ProductProcessor productProcessor = (ProductProcessor) baseForm
				.getBusinessObjectFactory().createObject(
						"productProcessorImplementer");
		Product product =  baseForm.getBusinessObjectFactory()
				.createObject(Product.class);

		// Look up the product using the seoCode first
		if (useSEOCodeForProductURLs != null
				&& useSEOCodeForProductURLs.equals("1")) {
			BeanUtils.copyProperties(product, baseForm);
			product.setSeoCode(product.getCode());
			product = productProcessor.getProductFromSeoCode(product);
		}

		// If that fails, look it up using the code
		if (product.getProductID() == 0) {
			BeanUtils.copyProperties(product, baseForm);
			product = productProcessor.getProductFromCode(product);
		}

		// If that fails, it's an error
		if (product.getProductID() == 0) {
			baseForm.getErrors().add(
					ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.productNotFound", product
							.getCode()));
			saveErrors(baseForm.getRequest(),
					(ActionMessages) baseForm.getErrors());
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
			return mapping.findForward("failure");
		}

		// Load all attribute-only SKUs
		String trackInventory = (String) baseForm.getSettings().getValue(
				"trackInventory");
		String applyDiscounts = (String) baseForm.getSettings().getValue(
				"applyDiscounts");
		if ((trackInventory != null && trackInventory.equals("1"))
				|| (applyDiscounts != null && applyDiscounts.equals("1"))) {
			Map attributeSKUs = productProcessor.loadAttributeSKUs();
			if (log.isDebugEnabled())
				log.debug("Atribute-only skus: " + attributeSKUs);
			request.setAttribute("attributeSKUs", attributeSKUs);
			Collection attributeSKUCollection = productProcessor
					.loadAttributeSKUCollection();
			request.setAttribute("attributeSKUCollection",
					attributeSKUCollection);
		}

		if (product.getProductSettings() != null) {

			// If the product is set to display attributes in a matrix, load the
			// matrix
			ProductSetting displayAsMatrix = (ProductSetting) product
					.getProductSettings().get("displayAsMatrix");
			if (displayAsMatrix != null && displayAsMatrix.getValue() != null
					&& displayAsMatrix.getValue().equals("1")) {
				Collection matrix = productProcessor.buildMatrix(product);
				if (matrix != null && !matrix.isEmpty()) {
					request.setAttribute("matrix", matrix);
				}
			}

			// Set the displayAdditionalImages variable
			// Set the displayAdditionalImages variable
			boolean displayAdditionalImages = false;
			boolean smallImageExists = false;
			boolean largeImageExists = false;
			ArrayList keys = new ArrayList(product.getProductSettings()
					.keySet());
			Iterator i = keys.iterator();
			while (i.hasNext()) {
				ProductSetting ps = (ProductSetting) product
						.getProductSettings().get((String) i.next());
				if ("IMAGE".equals(ps.getType())
						&& "SMALL".equals(ps.getName())) {
					smallImageExists = true;
				}
				if ("IMAGE".equals(ps.getType())
						&& "LARGE".equals(ps.getName())) {
					largeImageExists = true;
				}
				if (smallImageExists && largeImageExists) {
					displayAdditionalImages = true;
					break;
				}
			}
			request.setAttribute("displayAdditionalImages", new Boolean(
					displayAdditionalImages));

			// Set the displayAdditionalImages variable
			boolean displayIframesVideos = false;
			keys = new ArrayList(product.getProductSettings().keySet());
			Iterator j = keys.iterator();
			while (j.hasNext()) {
				ProductSetting ps = (ProductSetting) product.getProductSettings().get((String) j.next());
				if ("VIDEO".equals(ps.getType())) {
					displayIframesVideos = true;
					break;
				}
			}
			request.setAttribute("displayIframesVideos", new Boolean(displayIframesVideos));

			// Set the displayProductFacets variable
			boolean displayProductFacets = false;
			i = keys.iterator();
			while (i.hasNext()) {
				ProductSetting ps = (ProductSetting) product.getProductSettings().get((String) i.next());
				if (ps == null) {
					continue;
				}
				if ("FACET".equals(ps.getType())) {
					displayProductFacets = true;
					break;
				}
			}
			request.setAttribute("displayProductFacets", new Boolean(displayProductFacets));
		}

		// Load the product reviews and get an average based on 1-5
		if (baseForm.getSettings().getValue("enableProductReviews").equals("1")) {
		Collection productReviews = product.getProductReviews();
		double rating = 0.00;
		int countApprovedReviews = 0;
		for (Iterator iter = productReviews.iterator(); iter.hasNext();) {
			ProductReview pr = (ProductReview) iter.next();
			if (pr.getStatus() != null && pr.getStatus().intValue() == 1) {
				countApprovedReviews++;
				rating = rating + pr.getRating().doubleValue();
			}
		}
		if (countApprovedReviews > 0) {
			double average = rating / countApprovedReviews;
			double f = 0.5;
			double roundedAverage = f * Math.round(average / f);
			long iPart = (long) roundedAverage;
			double fPart = roundedAverage - iPart;
			request.setAttribute("countApprovedReviews", new Integer(
					countApprovedReviews));
			request.setAttribute("productReviews", productReviews);
			request.setAttribute("productReviewAverage", new Double(roundedAverage));
			request.setAttribute("wholeStars", new Double(iPart));
			request.setAttribute("halfStars", new Double(fPart));
		}
		}

		// Make sure a currentCategory is loaded for breadcrumbing
		synchronized (request.getSession()) {
			String categoryCode = (String) baseForm.get("category");
			Category currentCategory = (Category) request.getSession()
					.getAttribute("currentCategory");
			// Links coming from category page are kosher
			if (currentCategory != null && categoryCode != null
					&& currentCategory.getCode().equals(categoryCode)) {
			} else {
				// Links coming from a search result or other source:
				// Remove the currentCategory, then attempt to
				// populate it with the primary category of the product
				request.getSession().removeAttribute("currentCategory");
				if ((categoryCode == null || categoryCode.length() == 0)
						&& product.getPrimaryCategory() != null) {
					categoryCode = product.getPrimaryCategory().getCode();
				}
				if (categoryCode != null && categoryCode.length() > 0) {

					// Load the category
					Category category = (Category) baseForm.getBusinessObjectFactory().createObject(
							"categoryImplementer");
					category.setCode(categoryCode);
					category = productProcessor.getCategoryFromCode(category, false);
					if (category.getCategoryID() != 0) {
						request.getSession().setAttribute("currentCategory", category);
						request.getSession().setAttribute("categorySettings", category.getCategorySettings());
					}
				}
			}
		}

		// Add the product to the session's viewedProducts collection
		synchronized (request.getSession()) {
			Collection viewedProducts = (Collection) request.getSession()
					.getAttribute("viewedProducts");
			if (viewedProducts == null) {
				viewedProducts = new ArrayList();
				request.getSession().setAttribute("viewedProducts",
						viewedProducts);
			}
			Iterator i = viewedProducts.iterator();
			while (i.hasNext()) {
				Product p = (Product) i.next();
				if (p.getCode().equals(product.getCode())) {
					i.remove();
				}
			}
			viewedProducts.add(product);
		}

		request.setAttribute("product", product);
		request.setAttribute("productForm", product);
		request.setAttribute("productSettings", product.getProductSettings());

		// Set various meta tag fields - title tag first
		String pageName = "";
		if (product.getName() != null) {
			String parsedName = product.getName().replaceAll("\\<.*?>", "");
			if (parsedName.length() > 0) {
				pageName = parsedName;
			}
		}
		if (product.getProductSettings() != null) {
			ProductSetting ps = (ProductSetting) product.getProductSettings()
					.get("seoTitleTag");
			if (ps != null && ps.getValue() != null
					&& ps.getValue().length() > 0) {
				pageName = ps.getValue();
			}
		}
		if (pageName.length() > 0) {
			request.setAttribute("htmlHeadTitle",
					request.getAttribute("htmlHeadTitle") + " - " + pageName);
		}
		if (product.getProductSettings() != null) {
			ProductSetting ps = (ProductSetting) product.getProductSettings()
					.get("seoAppendToTitleTag");
			if (ps != null && ps.getValue() != null
					&& ps.getValue().equals("0")) {
				request.setAttribute("htmlHeadTitle", pageName);
				request.setAttribute("dontAppendToTitleTag", new Boolean(true));
			}
		}

		// Set the meta/keywords tag
		if (product.getKeywords() != null && product.getKeywords().length() > 0) {
			request.setAttribute("htmlHeadKeywords", product.getKeywords());
		}

		// Set the meta/description tag
		boolean descriptionSet = false;
		if (product.getProductSettings() != null) {
			ProductSetting ps = (ProductSetting) product.getProductSettings()
					.get("seoMetaDescription");
			if (ps != null && ps.getValue() != null
					&& ps.getValue().length() > 0) {
				request.setAttribute("htmlHeadDescription", ps.getValue());
				descriptionSet = true;
			}
		}
		if (!descriptionSet && product.getShortDescription() != null) {
			String desc = product.getShortDescription().replaceAll("\\<.*?>",
					"");
			if (desc.length() > 0) {
				request.setAttribute("htmlHeadDescription", desc);
				descriptionSet = true;
			}
		}
		if (!descriptionSet && product.getDescription() != null) {
			String desc = product.getDescription().replaceAll("\\<.*?>", "");
			if (desc.length() > 0) {
				request.setAttribute("htmlHeadDescription", desc);
			}
		}

		// Set the forward based on the pageLayoutStyle setting - full layout or
		// sidebar layout
		if (product.getProductSettings() != null) {
			ProductSetting ps = (ProductSetting) product.getProductSettings()
					.get("pageLayoutStyle");
			if (ps != null && ps.getValue() != null
					&& ps.getValue().equals("layout")) {
				return mapping.findForward("successFullLayout");
			}
		}
		return mapping.findForward("success");
	}
}
